/*
 * Copyright 2015 Program of Computer Graphics, Cornell University
 *     580 Rhodes Hall
 *     Cornell University
 *     Ithaca NY 14853-3801
 * Web: http://www.graphics.cornell.edu/
 *
 */

//===============================================================
//
// heBRDF.cpp
//   A reference implementation of the wave-optics reflectance
//     model from
//
//    Xiao D. He, Kenneth E. Torrance, Francois X. Sillion,
//      and Donald P. Greenberg.
//    A comprehensive physical model for light reflection.
//    Computer Graphics 25(4) (SIGGRAPH '91 Proceedings)
//      pp 175-186, July 1991.
//
//  Implemented by Liang Peng
//  based on Steve Westin's original code.
//  improvement:
//     (1) Check K_func calculation in Equ(82), set an upper bound for K.
//     (2) Decoupled the spatial terms from the spatial/spectral coupled
//         terms in the BRDF calculation.
//
//  -- 28/9/97
// 04/07/98: Fix for proper ANSI scoping
// 04/22/98: Add check for divide by zero in Sigma_func()
// 04/26/00: Change extension to lowercase for heBRDF.h
// 04/27/00: Reactivate rough-surface asymptotic solution for D term,
//	     using 18.5 (empirically derived) for the threshold.
// 02/02/04: Fix Fresnel calculation for rho_dd: needs to use half
//	     the angle between incident and exitant directions
// 02/04/04: Use trigonometric identity to get cos(half_angle)
//
//===============================================================

#include <math.h>
#include "Vector3.h"
#include "heBRDF.h"
#include <stdio.h>

#ifndef M_PI
#define M_PI                3.14159265358979323846
#endif // M_PI

static const double ANGLE_EPS = 1.0e-5;	     // Angle less than this is zero
static const double STINY = 1.0e-4; // in micro
static const double tinyEpsilon = 1.0e-12;
static const double smallEpsilon = 1.0e-8;

// Needed on Windows, as library doesn't include it
#ifdef WIN32
extern "C" double erfc (double );
#endif // WIN32

inline double SQR(const double& x) {
  return (x*x);
}

// Return an arbitrary unit vector perpendicular to v:
Vector3 PerpUnitVector3(const Vector3 &v) {

  Vector3 u = Vector3(1,0,0);
  if ( dot(u,v)<tinyEpsilon )
    u.set(0,1,0);
  Vector3 w = cross(v,u);
  w.MakeUnitVector();
  return w;
}

// constructor:
heBRDF::heBRDF(const double sigma0_, const double tau_, // in micrometers
               const double *lambda_, int nLambda_, // in nanometers:
               const double *k1_, const double *k2_, // index of
					     // refraction/extinction
               const double *rho_ud_) {

  int i;

  sigma0 = sigma0_ ;
  tau = tau_ ;

  nLambda = nLambda_;
  lambda = new double [nLambda];
  k1 = new double [nLambda];
  k2 = new double [nLambda];
  rho_ud = new double [nLambda];

  for (i=0;i<nLambda;i++) {
    lambda[i] = lambda_[i] * 0.001;  // conversion from nanometer to micron
    k1[i] = k1_[i];
    k2[i] = k2_[i];
  }

  if (rho_ud_) {
    double dLambda = lambda[1] - lambda[0];
    double sc = 1.0/M_PI;
    for (int i=0;i<nLambda;i++) {
      rho_ud[i] = sc*rho_ud_[i];    // scale to the proper unit:
    }
  }
  else {
    for (int i=0;i<nLambda;i++) {
      rho_ud[i] = 0;
    }
  }
}

// destructor:
heBRDF::~heBRDF() {
  delete lambda;
  delete k1;
  delete k2;
  delete rho_ud;
}

// specular component: Equ(70)
double* heBRDF::rho_sp (const Vector3& w_i, const Vector3& w_r,
                        const Vector3& n, const double& dw_i) {

  double *Rho = new double [nLambda];

  Vector3 w_mirror = 2*dot(w_i, n) * n - w_i;

  if ( dot(w_mirror, w_r)>= cos(sqrt(dw_i/M_PI)) ) {

    double *Rho_s = rho_s(w_i, n);
    for (int i=0;i<nLambda;i++) {
      Rho[i] = Rho_s[i] / ( dot(w_i, n) * dw_i) ;
    }
    delete [] Rho_s;
  }
  else {
    for (int i=0;i<nLambda;i++) {
      Rho[i] = 0;
    }
  }
  return (Rho);
}

// total reflectance: Equ(69)
double* heBRDF::rho_tot(const Vector3& w_i, const Vector3& w_r,
                        const Vector3& n, const double& dw_i) {

  double *Rho_dd = rho_dd (w_i, w_r, n);
  double *Rho_sp = rho_sp (w_i, w_r, n, dw_i);

  double *Rho = new double [nLambda];
  for (int i=0;i<nLambda;i++) {
    Rho[i] = Rho_dd[i] + Rho_sp[i] + rho_ud[i];
  }

  delete [] Rho_dd;
  delete [] Rho_sp;

  return (Rho);
}

// specular coefcients:   Equ(73, 74)
double* heBRDF::rho_s (const Vector3& w_i, const Vector3& n) {

  double cos_theta_i = dot(w_i, n);
  double sin_theta_i = sqrt ( 1.0 - SQR(cos_theta_i) );
  double cot_theta_i = cos_theta_i / sin_theta_i;

  //--------------------------------------------------
  // spatial-only dependent terms:
  double S;
  if ( sin_theta_i > ANGLE_EPS )	     // SHW 04/08/98: Provide for
    S =	 SQR(S_func(cot_theta_i));	     // theta_i -> 0
  else
    S = 1.0;

  // sigma: Equ(80)

  double sigma;
  /*
  if ( sigma0 < STINY )
    sigma = sigma0;
  else
  */
  sigma = Sigma_func(cos_theta_i, cos_theta_i);

  double *rho = new double [nLambda];

  //----------------------------------------------------
  //Spectral-Spatial coupled terms:
  for (int i=0;i<nLambda;i++) { // spectral sampling:
    double g = SQR(2.0*M_PI*(sigma/lambda[i])*(cos_theta_i+cos_theta_i));
    double F = FresnelCoef(cos_theta_i, k1[i], k2[i]);
    rho[i] = S*F*exp(-g);
  }
  return (rho);
}

//---------------------------------------------------------
// Functions depending only on angular configuration:
//--------------------------------------------------------

// Shadowing function: Equ(24)
double
heBRDF::S_func(const double& cot_theta) const {

  // double tau_over_sigma = tau/sigma0;
  if ( sigma0 < STINY ) return 1.0;	     // SHW 04/08/98 Allow smooth
					     // surface
  double cotexp = cot_theta*tau/(sigma0*2);
  double erfy = erfc(cotexp);

  double capitalLambda = 0.5 *
    ( exp(-SQR(cotexp))/(sqrt(M_PI)*cotexp) - erfy );

  double S = (1.0 - 0.5*erfy) / (capitalLambda + 1.0);
  return S;
}


// Get Fresnel Coef from Index of refraction/extinction
//  |F|^2  : Equ(75)
double
heBRDF::FresnelCoef(const double& cos_theta_i, const double& k1_, const double& k2_) const {

  double c = cos_theta_i;
  double s = sqrt(1-SQR(c));
  double c2 = SQR(c);
  double s2 = SQR(s);
  double t = s/c;
  double st = s*t;
  double st2 = SQR(st);

  double B = SQR(k1_) - SQR(k2_)-s2;
  double A = sqrt(SQR(B)+4*SQR(k1_*k2_));

  double  a2 = 0.5*(A+B);
  double  b2 = 0.5*(A-B);
  double  a = sqrt(a2);

  double  Fs = (a2+b2-2*a*c+c2)/(a2+b2+2*a*c+c2);
  double  Fp = Fs * (a2+b2-2*a*st+st2)/(a2+b2+2*a*st+st2);

  double F = 0.5*(Fs+Fp);

  return (F);

}


// Routine to find effective surface roughness
// involves Newton-Raphson method for finding z0
// Equ (80)
#define MAXITER 100
#define ACCURACY 1e-8
#define f_ACCURACY 1e-8
double
heBRDF::Sigma_func (double &cos_theta_i,
                    double &cos_theta_r) const {

  if ( fabs(sigma0) < ACCURACY ) return 0.0;
  if ( cos_theta_i < ACCURACY || cos_theta_r < ACCURACY )
    return 0.0;				     // Avoid divide-by-zero
  // get K_i, K_r:
  double tan_theta_i = sqrt(1-SQR(cos_theta_i))/ cos_theta_i;
  double tan_theta_r = sqrt(1-SQR(cos_theta_r))/ cos_theta_r;

  double K = K_func(tan_theta_i) + K_func(tan_theta_r);

  if (K==0)   // trivial case
    return sigma0;

  // calculate z0/sigma0 <- x
  double x, expn, func, func_prev, df_dx, dx;
  double f0 = sqrt(1.0/(8.0*M_PI)) * K;

  // A carefully choosen initial value,
  // convergent with 15 iteration per solving when f0=0->10^5:
  if (f0<=1)
    x = f0;
  else
    x = sqrt(2*log(f0));

  func_prev = -1;
  int i;
  for (i=0; i<MAXITER; i++) {

    // Evaluate the function and its derivative
    expn = exp(0.5*x*x);
    func = x * expn - f0;
    df_dx = (1+x*x)*expn;

    // predict next position with Newton method:
    dx = func / df_dx;
    x -= dx;

    if ( (fabs(dx) < ACCURACY) && ( fabs(func-func_prev)<f_ACCURACY) )
      break;

    func_prev = func;
  }

  double eff_sigma = sigma0 / sqrt(1 + SQR(x));
  return eff_sigma;
}


// Equ: (82-83)
double heBRDF::K_func(double& tan_theta ) const {

  // double eps = 1.0e-4;  // 20 arc second
  double eps = 1.0e-6;
  double maxNum = 1.0e+12;  // overflow:
  double K;

  // if (0.5*tau/sigma0*cot_theta)>5, then K<tan_theta*E-12
  if (fabs(tan_theta)< eps)
    K = 0;
  else if (tan_theta< maxNum)
    K = tan_theta * erfc(0.5*tau/(sigma0*tan_theta));
  else
    K =  maxNum;

  return K;
}


// Geometry factor G: Equ(76)
double heBRDF::G_func(const Vector3& w_i,
                      const Vector3& w_r,
                      const Vector3& n) const {

  Vector3 k_i = -w_i;
  Vector3 k_r = w_r;

  if (w_i == w_r) {  // in case same direction:
    k_r = k_r + smallEpsilon*Vector3(1,1,1);
    k_r.MakeUnitVector();
  }

  Vector3 s_i = cross(k_i, n);
  if (s_i.length() > smallEpsilon)
    s_i.MakeUnitVector();
  else
    s_i = PerpUnitVector3(n);

  Vector3 s_r = cross(k_r, n);
  if (s_r.length() > smallEpsilon)
    s_r.MakeUnitVector();
  else
    s_r = PerpUnitVector3(n);

  // sure perpendicular:
  Vector3 p_i = cross(s_i, k_i);
  Vector3 p_r = cross(s_r, k_r);

  Vector3 v = k_r - k_i;
  double v_term = SQR(dot(v,v)/dot(v, n));

  Vector3 crossVec = cross(k_i, k_r);

  double dot_term = ( SQR(dot(s_r, k_i)) + SQR(dot(p_r, k_i)) )
    * ( SQR(dot(s_i, k_r)) + SQR(dot(p_i, k_r)) );


  double G = v_term * dot_term / SQR(dot(crossVec, crossVec));

  return (G);
}

// last term of D summation
#define Dsumlim 1.0e-12
// maximum terms in D summation; this avoids overflow of m! term */
#define Dsummax 1000000

// Eq. 78
double heBRDF::D_func(const double& g,
                      const double& tau_over_lambda,
                      const double& v_xy2) const {

  double D;

  if ( g > 18.5 ) {
    // Rough surface: use the approximation from Beckmann, 1963 edition,
    // section 5.3, equation 47, page 87.
    D = SQR(M_PI/2.0*tau_over_lambda) / g
      * exp ( -v_xy2 *SQR(2*M_PI*tau_over_lambda) / (4.0*g) );
  }
  else {
    // Compute as the sum of an infinite series. (Section 5.3, Equation 35,
    // Page 86 of Beckmann '63.)
    // The series is a little odd, in that magnitude of the terms tends
    // to increase, then decrease. We terminate when terms get really small
    // AND are diminishing.

    double v_xy_tau_2 = v_xy2 * SQR(tau_over_lambda*2*M_PI);

    double sum = 0.0;
    double m_fact = 1.0;
    double this_term = 0.0,
           last_term = -1.0;

    for ( int m=1;
	  ( this_term > Dsumlim ) || ( this_term > last_term );
	  m++ ) {

      // Multiply the factor by g/m to update the g**m/(m!) term. */
      m_fact *= g / (double)m;
      last_term = this_term;
      // exp(-709) is about the point of underflow in double precision
      // (result is 1e-308). If the argument to the exponential is more
      // negative than this, the answer will always be zero, and we can
      // avoid the function call.
      double exp_base = -g - v_xy_tau_2/(4.0*m);
      if ( exp_base > -750.0 ) {
	this_term = exp ( exp_base ) * m_fact / m;
	sum += this_term;
      }
      else
	this_term = 0.0;

    }

    D =  sum * SQR(0.5*M_PI*tau_over_lambda);

  }

  return (D);

}


// directional diffuse BRDF: Equ(71)
double* heBRDF::rho_dd (const Vector3& w_i, const Vector3& w_r,
                        const Vector3& n) {

  double cos_theta_i = dot(w_i, n);
  double sin_theta_i = sqrt ( 1.0 - SQR(cos_theta_i) );
  double cot_theta_i = cos_theta_i / sin_theta_i;

  double cos_theta_r = dot(w_r, n);
  double sin_theta_r = sqrt ( 1.0 - SQR(cos_theta_r) );
  double cot_theta_r = cos_theta_r / sin_theta_r;

  // Compute half the angle between incident and exitant directions,
  // by means of the trigonometric identity
  // cos(x/2) = +- sqrt((1+cos(x))/2).
  // Since we know that 0 <= x <= pi, we can neglect the negative
  // branch.
  double cos_half_angle = sqrt ( (1+dot(w_i,w_r)) / 2 );

  //--------------------------------------------------
  // spatial dependent terms:
  // spatial-only dependent terms:
  double S = 1.0;
  if ( sin_theta_i > ANGLE_EPS )	     // 04/08/98 SHW: Provide for
    S *= S_func(cot_theta_i);		     // theta -> 0
  if ( sin_theta_r > ANGLE_EPS )
    S *= S_func(cot_theta_r);

  // sigma: Equ(80)
  double sigma;

  /*
  if ( sigma0 < STINY )
    sigma = sigma0;
  else
  */
  sigma = Sigma_func(cos_theta_i, cos_theta_r);

  // G:
  double G = G_func(w_i, w_r, n);

  // spatial term:
  double spatial_term;
  if ( (G*S) ==0)
    spatial_term = 0;
  else
    spatial_term = G*S/(M_PI*cos_theta_i*cos_theta_r);

  // v_xy:
  Vector3 v = w_r + w_i;   // k_r - k_i
  double v_xy2 = dot(v,v) - SQR(dot(v, n));

 //----------------------------------------------------
  //Spectral-Spatial coupled terms:
  double g, F, D;

  double *rho = new double [nLambda];
  for (int i=0;i<nLambda;i++) { // spectral sampling:
    g = SQR(2*M_PI*(sigma/lambda[i])*(cos_theta_i+cos_theta_r));
    F = FresnelCoef(cos_half_angle, k1[i], k2[i]);
    D = D_func(g, (tau/lambda[i]), v_xy2);
    rho[i] = spatial_term * F * D;

  }

  return (rho);
}

