/*
 * Copyright 2015 Program of Computer Graphics, Cornell University
 *     580 Rhodes Hall
 *     Cornell University
 *     Ithaca NY 14853-3801
 * Web: http://www.graphics.cornell.edu/
 *
 */

//===============================================================
//
// heBRDF.cpp
//   A reference implementation of the wave-optics reflectance
//     model from
//
//    Xiao D. He, Kenneth E. Torrance, Francois X. Sillion,
//      and Donald P. Greenberg.
//    A comprehensive physical model for light reflection.
//    Computer Graphics 25(4) (SIGGRAPH '91 Proceedings)
//      pp 175-186, July 1991.
//
//  Implemented by Liang Peng
//  based on Steve Westin's original code.
//  improvement:
//     (1) Check K_func calculation in Equ(82), set an upper bound for K.
//     (2) Decoupled the spatial terms from the spatial/spectral coupled
//         terms in the BRDF calculation.
//
//  -- 28/9/97
// 04/07/98: Fix for proper ANSI scoping
// 04/22/98: Add check for divide by zero in Sigma_func()
// 04/26/00: Change extension to lowercase for heBRDF.h
// 04/27/00: Reactivate rough-surface asymptotic solution for D term,
//	     using 18.5 (empirically derived) for the threshold.
// 02/02/04: Fix Fresnel calculation for rho_dd: needs to use half
//	     the angle between incident and exitant directions
// 02/04/04: Use trigonometric identity to get cos(half_angle)
//
//===============================================================

#ifndef _he_BRDF_H_
#define _he_BRDF_H_

// only rho_s + rho_dd is implemented:
class heBRDF {
public:

  // constructor:
  heBRDF(const double sigma0_, const double tau_, // in micro
         const double *lambda_, int nLambda_,     // in nanonmeter:
         const double *k1_, const double *k2_,  // index of refraction/extinction
         const double *rho_ud_=0);

  // destructor:
  ~heBRDF();

  //----------------------------------------------------
  // BRDF computation:
  //
  // (1) I try to group the computation for spectral terms separately from
  // angular terms as much as possible to avoid redundant calculation.
  //
  // (2) I try to use dimensionless quantities as much as possible to
  // increase the precision in the computation:
  //

  // directional diffuse BRDF: Equ(71)
  // return a spectrum in rho:
  double *rho_dd (const Vector3& w_i, const Vector3& w_r, const Vector3& n);

  // specular component: Equ(70)
  double *rho_sp (const Vector3& w_i, const Vector3& w_r, const Vector3& n,
                  const double& dw_i);

  // total reflectance: Equ(69)
  double *rho_tot (const Vector3& w_i, const Vector3& w_r, const Vector3& n,
                   const double& dw_i);

  // specular coefficients:   Equ(73, 74)
  // Only used when the incident direction is within the solid angle Delta
  // return a spectrum in rho
  double *rho_s (const Vector3& w_i, const Vector3& n);

 private:

  //------------------------------------------------------
  // Functions depending only on angular configuration:

  double S_func(const double& cot_theta) const; // Eq. 24

  double Sigma_func (double &cos_theta_i,    // Eq. 80
                     double &cos_theta_r) const ;

  double K_func(double& cot_theta) const;    // Eq. 82, 83

  double G_func(const Vector3& w_i,          // Eq. 76
                const Vector3& w_r,
                const Vector3& n) const;

  //------------------------------------------------
  // Functions depending both on angles and on wavelength:

  double FresnelCoef(const double& cos_theta_i, // Eq. 75
                     const double& k1_,
                     const double& k2_) const ;

  double D_func(const double& g,             // Eq. 78
                const double& tau_over_lambda,
                const double& v_xy2) const;

  //------------------------------------------------
  // Data members:

  // RMS roughness and autocorrelation length
  double sigma0, tau;

  // Wavelength, real and imaginary parts of refractive index
  double *lambda, *k1, *k2;

  // Number of elements in arrays lambda, k1, k2
  int nLambda;

  // Uniform diffuse (Lambertian) term
  double *rho_ud;
};

#endif
